#include <libc64_dsp/include/inc_overlay.h>
//#include "scalers.h"
#include <stdint.h>

#define MAX_WIDTH 320

typedef struct __blendTable {
    uint32_t *table;
    uint32_t num;
} _blendTable;

#define yuv_diff(yuv1, yuv2) ((_cmpgtu4(_subabs4(yuv1, yuv2), 0x00300706))?1:0)


static void prepare_blend_table(_blendTable * restrict blendTable, uint32_t * restrict blendTables, int width)
{
    _nassert((int)blendTable % 8 == 0);
    _nassert((int)blendTables % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);

    unsigned int i;

    for (i = 0; i < 15; i++)
    {
        blendTable[i].table = &(blendTables[i*width]);
    }
}

static void prepare_yuv_palette(const uint16_t * restrict RGBpal, uint32_t * restrict YUVpal)
{
    _nassert((int)RGBpal % 8 == 0);
    _nassert((int)YUVpal % 8 == 0);

    unsigned int i;

    for (i = 0; i < 256; i++)
    {
        // original hqx code:
        uint32_t r, g, b;
        uint32_t y, u, v;

        r = (i << 3) & (31 << 3);
        g = (i >> 3) & (63 << 2);
        b = (i >> 8) & (31 << 3);

        r |= r >> 5;
        g |= g >> 6;
        b |= b >> 5;

        //y = (uint32_t)(0.299*r + 0.587*g + 0.114*b);
        //u = (uint32_t)(-0.169*r - 0.331*g + 0.5*b) + 128;
        //v = (uint32_t)(0.5*r - 0.419*g - 0.081*b) + 128;

        //y = ( 0.299dd   *r + 0.587dd   *g + 0.114dd   *b);
        //u = (-0.168736dd*r - 0.331264dd*g + 0.5dd     *b) + 128.0dd;
        //v = ( 0.5dd     *r - 0.418688dd*g - 0.081312dd*b) + 128.0dd;

        y = ((((uint32_t)( 2508194*r + 4924113*g +  956302*b))             ) >> 23);
        //u = ((((uint32_t)(-1415460*r - 2778844*g + 4194304*b)) + 1073741824) >> 23);
        u = ((((uint32_t)( 4194304*b - 1415460*r - 2778844*g)) + 1073741824) >> 23);
        v = ((((uint32_t)( 4194304*r - 3512210*g -  682094*b)) + 1073741824) >> 23);

        // this(>>18) works for rgb888 values created from rgb565
        //y = ((((uint32_t)(  78381*r + 153879*g +  29885*b))          ) >> 18);
        // this(>>19) works for all rgb888 values
        //y = ((((uint32_t)( 156763*r + 307758*g +  59769*b))          ) >> 19);
        //u = ((((uint32_t)( -11058*r -  21710*g +  32768*b)) + 8388608) >> 16);
        //v = ((((uint32_t)(  32768*r -  27439*g -   5329*b)) + 8388608) >> 16);

        YUVpal[i] = (y << 16) + (u << 8) + v;
    }
}

static void convert_yuv_rgb_8(const uint8_t * restrict src, uint32_t * restrict yuvdst, uint16_t * restrict rgbdst, const uint32_t * restrict YUVpal, const uint16_t * restrict RGBpal, int width)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)yuvdst % 8 == 0);
    _nassert((int)rgbdst % 8 == 0);
    _nassert((int)YUVpal % 8 == 0);
    _nassert((int)RGBpal % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);

    int i;

    for (i = 0; i < width; i++)
    {
        yuvdst[i] = YUVpal[src[i]];
        rgbdst[i] = RGBpal[src[i]];
    }

    yuvdst[-1] = yuvdst[0];
    yuvdst[width] = yuvdst[width - 1];

    rgbdst[-1] = rgbdst[0];
    rgbdst[width] = rgbdst[width - 1];
}

static void convert_yuv_16(uint16_t * restrict src, uint32_t * restrict dst, int width)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);

    int i;

    src[-1] = src[0];
    src[width] = src[width - 1];

    for (i = 0; i < width; i++)
    {
        //uint8_t R, G, B;
        uint32_t r, g, b;
        uint32_t y, u, v;

        //R = (src[i] >>  0) & 31;
        //G = (src[i] >>  5) & 63;
        //B = (src[i] >> 11) & 31;

        //r = (R << 3) | (R >> 2);
        //g = (G << 2) | (G >> 4);
        //b = (B << 3) | (B >> 2);

        r = (src[i] << 3) & (31 << 3);
        g = (src[i] >> 3) & (63 << 2);
        b = (src[i] >> 8) & (31 << 3);

        r |= r >> 5;
        g |= g >> 6;
        b |= b >> 5;

        //y = (uint32_t)(0.299*r + 0.587*g + 0.114*b);
        //u = (uint32_t)(-0.169*r - 0.331*g + 0.5*b) + 128;
        //v = (uint32_t)(0.5*r - 0.419*g - 0.081*b) + 128;

        //y = ( 0.299dd   *r + 0.587dd   *g + 0.114dd   *b);
        //u = (-0.168736dd*r - 0.331264dd*g + 0.5dd     *b) + 128.0dd;
        //v = ( 0.5dd     *r - 0.418688dd*g - 0.081312dd*b) + 128.0dd;

        y = ((((uint32_t)( 2508194*r + 4924113*g +  956302*b))             ) >> 23);
        //u = ((((uint32_t)(-1415460*r - 2778844*g + 4194304*b)) + 1073741824) >> 23);
        u = ((((uint32_t)( 4194304*b - 1415460*r - 2778844*g)) + 1073741824) >> 23);
        v = ((((uint32_t)( 4194304*r - 3512210*g -  682094*b)) + 1073741824) >> 23);

        // this(>>18) works for rgb888 values created from rgb565
        //y = ((((uint32_t)(  78381*r + 153879*g +  29885*b))          ) >> 18);
        // this(>>19) works for all rgb888 values
        //y = ((((uint32_t)( 156763*r + 307758*g +  59769*b))          ) >> 19);
        //u = ((((uint32_t)( -11058*r -  21710*g +  32768*b)) + 8388608) >> 16);
        //v = ((((uint32_t)(  32768*r -  27439*g -   5329*b)) + 8388608) >> 16);

        dst[i] = (y << 16) + (u << 8) + v;
    }

    dst[-1] = dst[0];
    dst[width] = dst[width - 1];
}

static void calculate_pattern(const uint32_t * restrict yuvsrc1, const uint32_t * restrict yuvsrc2, const uint32_t * restrict yuvsrc3, uint8_t * restrict dst, int width)
{
    _nassert((int)yuvsrc1 % 8 == 0);
    _nassert((int)yuvsrc2 % 8 == 0);
    _nassert((int)yuvsrc3 % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);

    uint32_t e;
    uint8_t pattern;

    // before first
    {
        e = yuvsrc2[-1];
        pattern  = yuv_diff(e, yuvsrc1[-1]) << 0;
        pattern |= yuv_diff(e, yuvsrc1[-1]) << 1;
        pattern |= yuv_diff(e, yuvsrc1[ 0]) << 2;
        pattern |= yuv_diff(e, yuvsrc2[-1]) << 3;
        pattern |= yuv_diff(e, yuvsrc2[ 0]) << 4;
        pattern |= yuv_diff(e, yuvsrc3[-1]) << 5;
        pattern |= yuv_diff(e, yuvsrc3[-1]) << 6;
        pattern |= yuv_diff(e, yuvsrc3[ 0]) << 7;

        dst[-1] = pattern;
    }

    for (; width != 0; width--)
    {
        e = yuvsrc2[0];
        pattern  = yuv_diff(e, yuvsrc1[-1]) << 0;
        pattern |= yuv_diff(e, yuvsrc1[ 0]) << 1;
        pattern |= yuv_diff(e, yuvsrc1[ 1]) << 2;
        pattern |= yuv_diff(e, yuvsrc2[-1]) << 3;
        pattern |= yuv_diff(e, yuvsrc2[ 1]) << 4;
        pattern |= yuv_diff(e, yuvsrc3[-1]) << 5;
        pattern |= yuv_diff(e, yuvsrc3[ 0]) << 6;
        pattern |= yuv_diff(e, yuvsrc3[ 1]) << 7;

        *dst = pattern;

        yuvsrc1++;
        yuvsrc2++;
        yuvsrc3++;
        dst++;
    }

    // after last
    {
        e = yuvsrc2[0];
        pattern  = yuv_diff(e, yuvsrc1[-1]) << 0;
        pattern |= yuv_diff(e, yuvsrc1[ 0]) << 1;
        pattern |= yuv_diff(e, yuvsrc1[ 0]) << 2;
        pattern |= yuv_diff(e, yuvsrc2[-1]) << 3;
        pattern |= yuv_diff(e, yuvsrc2[ 0]) << 4;
        pattern |= yuv_diff(e, yuvsrc3[-1]) << 5;
        pattern |= yuv_diff(e, yuvsrc3[ 0]) << 6;
        pattern |= yuv_diff(e, yuvsrc3[ 0]) << 7;

        *dst = pattern;
    }
}

static void inline fill_blendtable_helper(const uint8_t * restrict patternTable, const uint16_t * restrict hqTable, _blendTable *blendTable, int width)
{
    _nassert((int)patternTable % 8 == 0);
    _nassert((int)hqTable % 8 == 0);
    _nassert((int)blendTable % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);

    int index;

    for (index = 0; index < 15; index++)
    {
        blendTable[index].num = 0;
    }

    for (index = 0; index < width; index++)
    {
        uint32_t hqValue, rule, ruleIndex;
        uint32_t * restrict table;

        hqValue = hqTable[patternTable[index]];
        rule = hqValue & 0xff;

        ruleIndex = blendTable[rule].num;
        blendTable[rule].num = ruleIndex + 1;

        table = blendTable[rule].table;
        table[ruleIndex] = index | ((hqValue & 0xff00) << 8);
    }
}

static void fill_blendtable1(const uint8_t * restrict patternTable, const uint16_t * restrict hqTable, _blendTable *blendTable, int width)
{
    _nassert((int)patternTable % 8 == 0);
    _nassert((int)hqTable % 8 == 0);
    _nassert((int)blendTable % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);

    int counter;
    uint32_t *blendPart;
    uint32_t blendValue, rule, ruleIndex;
    int32_t index;
    uint32_t * restrict table;

    fill_blendtable_helper(patternTable, hqTable, blendTable, width);

#define B yuvsrc1[index]
#define D yuvsrc2[index-1]
#define F yuvsrc2[index+1]
#define H yuvsrc3[index]

    // todo: can use pattern table (+- 1) instead of yuv_diff

    counter = blendTable[12].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[12].table; counter != 0; counter--)
        {
            blendValue = *blendPart;
            blendPart++;

            index = blendValue & 0xffff;
            //rule = (blendValue >> (yuv_diff(B, D)?20:16)) & 0x0f;
            rule = (blendValue >> ((patternTable[index-1] & (1 << 2))?20:16)) & 0x0f;

            ruleIndex = blendTable[rule].num;
            blendTable[rule].num = ruleIndex + 1;

            table = blendTable[rule].table;
            table[ruleIndex] = index;
        }
    }

    counter = blendTable[13].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[13].table; counter != 0; counter--)
        {
            blendValue = *blendPart;
            blendPart++;

            index = blendValue & 0xffff;
            //rule = (blendValue >> (yuv_diff(B, F)?20:16)) & 0x0f;
            rule = (blendValue >> ((patternTable[index+1] & (1 << 0))?20:16)) & 0x0f;

            ruleIndex = blendTable[rule].num;
            blendTable[rule].num = ruleIndex + 1;

            table = blendTable[rule].table;
            table[ruleIndex] = index;
        }
    }

    counter = blendTable[14].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[14].table; counter != 0; counter--)
        {
            blendValue = *blendPart;
            blendPart++;

            index = blendValue & 0xffff;
            //rule = (blendValue >> (yuv_diff(D, H)?20:16)) & 0x0f;
            rule = (blendValue >> ((patternTable[index-1] & (1 << 7))?20:16)) & 0x0f;

            ruleIndex = blendTable[rule].num;
            blendTable[rule].num = ruleIndex + 1;

            table = blendTable[rule].table;
            table[ruleIndex] = index;
        }
    }

#undef B
#undef D
#undef F
#undef H
}

static void fill_blendtable2(const uint8_t * restrict patternTable, const uint16_t * restrict hqTable, _blendTable *blendTable, int width)
{
    _nassert((int)patternTable % 8 == 0);
    _nassert((int)hqTable % 8 == 0);
    _nassert((int)blendTable % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);

    int counter;
    uint32_t *blendPart;
    uint32_t blendValue, rule, ruleIndex;
    int32_t index;
    uint32_t * restrict table;

    fill_blendtable_helper(patternTable, hqTable, blendTable, width);

#define B yuvsrc2[index+1]
#define D yuvsrc1[index]
#define F yuvsrc3[index]
#define H yuvsrc2[index-1]

    // todo: can use pattern table (+- 1) instead of yuv_diff

    counter = blendTable[12].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[12].table; counter != 0; counter--)
        {
            blendValue = *blendPart;
            blendPart++;

            index = blendValue & 0xffff;
            //rule = (blendValue >> (yuv_diff(B, D)?20:16)) & 0x0f;
            rule = (blendValue >> ((patternTable[index+1] & (1 << 0))?20:16)) & 0x0f;

            ruleIndex = blendTable[rule].num;
            blendTable[rule].num = ruleIndex + 1;

            table = blendTable[rule].table;
            table[ruleIndex] = index;
        }
    }

    counter = blendTable[13].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[13].table; counter != 0; counter--)
        {
            blendValue = *blendPart;
            blendPart++;

            index = blendValue & 0xffff;
            //rule = (blendValue >> (yuv_diff(B, F)?20:16)) & 0x0f;
            rule = (blendValue >> ((patternTable[index+1] & (1 << 5))?20:16)) & 0x0f;

            ruleIndex = blendTable[rule].num;
            blendTable[rule].num = ruleIndex + 1;

            table = blendTable[rule].table;
            table[ruleIndex] = index;
        }
    }

    counter = blendTable[14].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[14].table; counter != 0; counter--)
        {
            blendValue = *blendPart;
            blendPart++;

            index = blendValue & 0xffff;
            //rule = (blendValue >> (yuv_diff(D, H)?20:16)) & 0x0f;
            rule = (blendValue >> ((patternTable[index-1] & (1 << 2))?20:16)) & 0x0f;

            ruleIndex = blendTable[rule].num;
            blendTable[rule].num = ruleIndex + 1;

            table = blendTable[rule].table;
            table[ruleIndex] = index;
        }
    }

#undef B
#undef D
#undef F
#undef H
}

static void fill_blendtable3(const uint8_t * restrict patternTable, const uint16_t * restrict hqTable, _blendTable *blendTable, int width)
{
    _nassert((int)patternTable % 8 == 0);
    _nassert((int)hqTable % 8 == 0);
    _nassert((int)blendTable % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);

    int counter;
    uint32_t *blendPart;
    uint32_t blendValue, rule, ruleIndex;
    int32_t index;
    uint32_t * restrict table;

    fill_blendtable_helper(patternTable, hqTable, blendTable, width);

#define B yuvsrc3[index]
#define D yuvsrc2[index+1]
#define F yuvsrc2[index-1]
#define H yuvsrc1[index]

    // todo: can use pattern table (+- 1) instead of yuv_diff

    counter = blendTable[12].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[12].table; counter != 0; counter--)
        {
            blendValue = *blendPart;
            blendPart++;

            index = blendValue & 0xffff;
            //rule = (blendValue >> (yuv_diff(B, D)?20:16)) & 0x0f;
            rule = (blendValue >> ((patternTable[index+1] & (1 << 5))?20:16)) & 0x0f;

            ruleIndex = blendTable[rule].num;
            blendTable[rule].num = ruleIndex + 1;

            table = blendTable[rule].table;
            table[ruleIndex] = index;
        }
    }

    counter = blendTable[13].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[13].table; counter != 0; counter--)
        {
            blendValue = *blendPart;
            blendPart++;

            index = blendValue & 0xffff;
            //rule = (blendValue >> (yuv_diff(B, F)?20:16)) & 0x0f;
            rule = (blendValue >> ((patternTable[index-1] & (1 << 7))?20:16)) & 0x0f;

            ruleIndex = blendTable[rule].num;
            blendTable[rule].num = ruleIndex + 1;

            table = blendTable[rule].table;
            table[ruleIndex] = index;
        }
    }

    counter = blendTable[14].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[14].table; counter != 0; counter--)
        {
            blendValue = *blendPart;
            blendPart++;

            index = blendValue & 0xffff;
            //rule = (blendValue >> (yuv_diff(D, H)?20:16)) & 0x0f;
            rule = (blendValue >> ((patternTable[index+1] & (1 << 0))?20:16)) & 0x0f;

            ruleIndex = blendTable[rule].num;
            blendTable[rule].num = ruleIndex + 1;

            table = blendTable[rule].table;
            table[ruleIndex] = index;
        }
    }

#undef B
#undef D
#undef F
#undef H
}

static void fill_blendtable4(const uint8_t * restrict patternTable, const uint16_t * restrict hqTable, _blendTable *blendTable, int width)
{
    _nassert((int)patternTable % 8 == 0);
    _nassert((int)hqTable % 8 == 0);
    _nassert((int)blendTable % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);

    int counter;
    uint32_t *blendPart;
    uint32_t blendValue, rule, ruleIndex;
    int32_t index;
    uint32_t * restrict table;

    fill_blendtable_helper(patternTable, hqTable, blendTable, width);

#define B yuvsrc2[index-1]
#define D yuvsrc3[index]
#define F yuvsrc1[index]
#define H yuvsrc2[index+1]

    // todo: can use pattern table (+- 1) instead of yuv_diff

    counter = blendTable[12].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[12].table; counter != 0; counter--)
        {
            blendValue = *blendPart;
            blendPart++;

            index = blendValue & 0xffff;
            //rule = (blendValue >> (yuv_diff(B, D)?20:16)) & 0x0f;
            rule = (blendValue >> ((patternTable[index-1] & (1 << 7))?20:16)) & 0x0f;

            ruleIndex = blendTable[rule].num;
            blendTable[rule].num = ruleIndex + 1;

            table = blendTable[rule].table;
            table[ruleIndex] = index;
        }
    }

    counter = blendTable[13].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[13].table; counter != 0; counter--)
        {
            blendValue = *blendPart;
            blendPart++;

            index = blendValue & 0xffff;
            //rule = (blendValue >> (yuv_diff(B, F)?20:16)) & 0x0f;
            rule = (blendValue >> ((patternTable[index-1] & (1 << 2))?20:16)) & 0x0f;

            ruleIndex = blendTable[rule].num;
            blendTable[rule].num = ruleIndex + 1;

            table = blendTable[rule].table;
            table[ruleIndex] = index;
        }
    }

    counter = blendTable[14].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[14].table; counter != 0; counter--)
        {
            blendValue = *blendPart;
            blendPart++;

            index = blendValue & 0xffff;
            //rule = (blendValue >> (yuv_diff(D, H)?20:16)) & 0x0f;
            rule = (blendValue >> ((patternTable[index+1] & (1 << 5))?20:16)) & 0x0f;

            ruleIndex = blendTable[rule].num;
            blendTable[rule].num = ruleIndex + 1;

            table = blendTable[rule].table;
            table[ruleIndex] = index;
        }
    }

#undef B
#undef D
#undef F
#undef H
}

#define GROW(n) n = (n | n << 16) & 0x07e0f81f
static uint16_t inline pack(uint32_t n) { n &= 0x07e0f81f; return n | (n >> 16); }

static uint16_t inline blend1(uint32_t A, uint32_t B) {
  GROW(A); GROW(B);
  A = (A * 3 + B) >> 2;
  return pack(A);
}

static uint16_t inline blend2(uint32_t A, uint32_t B, uint32_t C) {
  GROW(A); GROW(B); GROW(C);
  return pack((A * 2 + B + C) >> 2);
}

static uint16_t inline blend3(uint32_t A, uint32_t B, uint32_t C) {
  GROW(A); GROW(B); GROW(C);
  return pack((A * 5 + B * 2 + C) >> 3);
}

static uint16_t inline blend4(uint32_t A, uint32_t B, uint32_t C) {
  GROW(A); GROW(B); GROW(C);
  return pack((A * 6 + B + C) >> 3);
}

static uint16_t inline blend5(uint32_t A, uint32_t B, uint32_t C) {
  GROW(A); GROW(B); GROW(C);
  return pack((A * 2 + (B + C) * 3) >> 3);
}

static uint16_t inline blend6(uint32_t A, uint32_t B, uint32_t C) {
  GROW(A); GROW(B); GROW(C);
  return pack((A * 14 + B + C) >> 4);
}

static void fill_dst1(const uint16_t * restrict rgbsrc1, const uint16_t * restrict rgbsrc2, const uint16_t * restrict rgbsrc3, uint16_t * restrict dst, _blendTable *blendTable)
{
    _nassert((int)rgbsrc1 % 8 == 0);
    _nassert((int)rgbsrc2 % 8 == 0);
    _nassert((int)rgbsrc3 % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert((int)blendTable % 8 == 0);

    int counter;
    uint32_t *blendPart;
    int32_t index;

#define E rgbsrc2[index]
#define A rgbsrc1[index-1]
#define B rgbsrc1[index]
#define D rgbsrc2[index-1]
#define F rgbsrc2[index+1]
#define H rgbsrc3[index]

    counter = blendTable[0].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[0].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = E;
        }
    }

    counter = blendTable[1].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[1].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend1(E, A);
        }
    }

    counter = blendTable[2].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[2].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend1(E, D);
        }
    }

    counter = blendTable[3].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[3].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend1(E, B);
        }
    }

    counter = blendTable[4].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[4].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend2(E, D, B);
        }
    }

    counter = blendTable[5].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[5].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend2(E, A, B);
        }
    }

    counter = blendTable[6].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[6].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend2(E, A, D);
        }
    }

    counter = blendTable[7].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[7].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend3(E, B, D);
        }
    }

    counter = blendTable[8].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[8].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend3(E, D, B);
        }
    }

    counter = blendTable[9].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[9].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend4(E, D, B);
        }
    }

    counter = blendTable[10].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[10].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend5(E, D, B);
        }
    }

    counter = blendTable[11].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[11].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend6(E, D, B);
        }
    }

#undef E
#undef A
#undef B
#undef D
#undef F
#undef H
}

static void fill_dst2(const uint16_t * restrict rgbsrc1, const uint16_t * restrict rgbsrc2, const uint16_t * restrict rgbsrc3, uint16_t * restrict dst, _blendTable *blendTable)
{
    _nassert((int)rgbsrc1 % 8 == 0);
    _nassert((int)rgbsrc2 % 8 == 0);
    _nassert((int)rgbsrc3 % 8 == 0);
    _nassert((int)dst % 2 == 0);
    _nassert((int)(dst-1) % 8 == 0);
    _nassert((int)blendTable % 8 == 0);

    int counter;
    uint32_t *blendPart;
    int32_t index;

#define E rgbsrc2[index]
#define A rgbsrc1[index+1]
#define B rgbsrc2[index+1]
#define D rgbsrc1[index]
#define F rgbsrc3[index]
#define H rgbsrc2[index-1]

    counter = blendTable[0].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[0].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = E;
        }
    }

    counter = blendTable[1].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[1].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend1(E, A);
        }
    }

    counter = blendTable[2].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[2].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend1(E, D);
        }
    }

    counter = blendTable[3].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[3].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend1(E, B);
        }
    }

    counter = blendTable[4].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[4].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend2(E, D, B);
        }
    }

    counter = blendTable[5].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[5].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend2(E, A, B);
        }
    }

    counter = blendTable[6].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[6].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend2(E, A, D);
        }
    }

    counter = blendTable[7].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[7].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend3(E, B, D);
        }
    }

    counter = blendTable[8].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[8].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend3(E, D, B);
        }
    }

    counter = blendTable[9].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[9].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend4(E, D, B);
        }
    }

    counter = blendTable[10].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[10].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend5(E, D, B);
        }
    }

    counter = blendTable[11].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[11].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend6(E, D, B);
        }
    }

#undef E
#undef A
#undef B
#undef D
#undef F
#undef H
}

static void fill_dst3(const uint16_t * restrict rgbsrc1, const uint16_t * restrict rgbsrc2, const uint16_t * restrict rgbsrc3, uint16_t * restrict dst, _blendTable *blendTable)
{
    _nassert((int)rgbsrc1 % 8 == 0);
    _nassert((int)rgbsrc2 % 8 == 0);
    _nassert((int)rgbsrc3 % 8 == 0);
    _nassert((int)dst % 2 == 0);
    _nassert((int)(dst-1) % 8 == 0);
    _nassert((int)blendTable % 8 == 0);

    int counter;
    uint32_t *blendPart;
    int32_t index;

#define E rgbsrc2[index]
#define A rgbsrc3[index+1]
#define B rgbsrc3[index]
#define D rgbsrc2[index+1]
#define F rgbsrc2[index-1]
#define H rgbsrc1[index]

    counter = blendTable[0].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[0].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = E;
        }
    }

    counter = blendTable[1].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[1].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend1(E, A);
        }
    }

    counter = blendTable[2].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[2].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend1(E, D);
        }
    }

    counter = blendTable[3].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[3].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend1(E, B);
        }
    }

    counter = blendTable[4].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[4].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend2(E, D, B);
        }
    }

    counter = blendTable[5].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[5].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend2(E, A, B);
        }
    }

    counter = blendTable[6].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[6].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend2(E, A, D);
        }
    }

    counter = blendTable[7].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[7].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend3(E, B, D);
        }
    }

    counter = blendTable[8].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[8].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend3(E, D, B);
        }
    }

    counter = blendTable[9].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[9].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend4(E, D, B);
        }
    }

    counter = blendTable[10].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[10].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend5(E, D, B);
        }
    }

    counter = blendTable[11].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[11].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend6(E, D, B);
        }
    }

#undef E
#undef A
#undef B
#undef D
#undef F
#undef H
}

static void fill_dst4(const uint16_t * restrict rgbsrc1, const uint16_t * restrict rgbsrc2, const uint16_t * restrict rgbsrc3, uint16_t * restrict dst, _blendTable *blendTable)
{
    _nassert((int)rgbsrc1 % 8 == 0);
    _nassert((int)rgbsrc2 % 8 == 0);
    _nassert((int)rgbsrc3 % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert((int)blendTable % 8 == 0);

    int counter;
    uint32_t *blendPart;
    int32_t index;

#define E rgbsrc2[index]
#define A rgbsrc3[index-1]
#define B rgbsrc2[index-1]
#define D rgbsrc3[index]
#define F rgbsrc1[index]
#define H rgbsrc2[index+1]

    counter = blendTable[0].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[0].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = E;
        }
    }

    counter = blendTable[1].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[1].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend1(E, A);
        }
    }

    counter = blendTable[2].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[2].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend1(E, D);
        }
    }

    counter = blendTable[3].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[3].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend1(E, B);
        }
    }

    counter = blendTable[4].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[4].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend2(E, D, B);
        }
    }

    counter = blendTable[5].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[5].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend2(E, A, B);
        }
    }

    counter = blendTable[6].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[6].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend2(E, A, D);
        }
    }

    counter = blendTable[7].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[7].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend3(E, B, D);
        }
    }

    counter = blendTable[8].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[8].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend3(E, D, B);
        }
    }

    counter = blendTable[9].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[9].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend4(E, D, B);
        }
    }

    counter = blendTable[10].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[10].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend5(E, D, B);
        }
    }

    counter = blendTable[11].num;
    _nassert(counter <= MAX_WIDTH);
    if (counter)
    {
        for (blendPart = blendTable[11].table; counter != 0; counter--)
        {
            index = *blendPart;
            blendPart++;

            dst[2*index] = blend6(E, D, B);
        }
    }

#undef E
#undef A
#undef B
#undef D
#undef F
#undef H
}

void _hq2x_8_16(const uint8_t * restrict src, uint16_t * restrict dst, int width, int srcstride, int dststride, int height, uint16_t *hqTable, uint16_t *rgbpalette)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);
    _nassert(srcstride >= width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);
    _nassert(height >= 4);
    _nassert((int)hqTable % 8 == 0);
    _nassert((int)rgbpalette % 8 == 0);

    // hqTable and rgbpalette should be in L1D SRAM

    uint8_t * restrict tmpsrc1;
    uint8_t * restrict tmpsrc2;
    uint8_t * restrict tmpsrc3;
    uint8_t * restrict tmpsrc4;

    uint16_t * restrict rgbsrc1;
    uint16_t * restrict rgbsrc2;
    uint16_t * restrict rgbsrc3;
    uint16_t * restrict rgbsrc4;

    uint32_t * restrict yuvsrc1;
    uint32_t * restrict yuvsrc2;
    uint32_t * restrict yuvsrc3;
    uint32_t * restrict yuvsrc4;

    uint8_t * restrict patternsrc1;

    uint16_t * restrict tmpdst1;
    uint16_t * restrict tmpdst2;

    uint32_t * restrict yuvpalette;

    _blendTable * restrict blendTable;

    tmpsrc1 = (uint8_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP); // tmpsrc1 is aligned to 8

    // start copying first line to tmpsrc1
    syscalls.qdma_copy1d(QDMA_0, tmpsrc1, src, width);

    tmpsrc2 = (uint8_t *) ( ((uintptr_t) tmpsrc1) + width); // tmpsrc2 is aligned to 8
    tmpsrc3 = (uint8_t *) ( ((uintptr_t) tmpsrc2) + width); // tmpsrc3 is aligned to 8
    tmpsrc4 = (uint8_t *) ( ((uintptr_t) tmpsrc3) + width); // tmpsrc4 is aligned to 8

    rgbsrc1 = (uint16_t *) ( ((uintptr_t) tmpsrc4) +     width + 8); // rgbsrc1 is aligned to 8 and has place for rgbsrc1[-1] and rgbsrc1[width]
    rgbsrc2 = (uint16_t *) ( ((uintptr_t) rgbsrc1) + 2 * width + 8); // rgbsrc2 is aligned to 8 and has place for rgbsrc2[-1] and rgbsrc2[width]
    rgbsrc3 = (uint16_t *) ( ((uintptr_t) rgbsrc2) + 2 * width + 8); // rgbsrc3 is aligned to 8 and has place for rgbsrc3[-1] and rgbsrc3[width]
    rgbsrc4 = (uint16_t *) ( ((uintptr_t) rgbsrc3) + 2 * width + 8); // rgbsrc4 is aligned to 8 and has place for rgbsrc4[-1] and rgbsrc4[width]

    yuvsrc1 = (uint32_t *) ( ((uintptr_t) rgbsrc4) + 2 * width + 8); // yuvsrc1 is aligned to 8 and has place for yuvsrc1[-1] and yuvsrc1[width]
    yuvsrc2 = (uint32_t *) ( ((uintptr_t) yuvsrc1) + 4 * width + 8); // yuvsrc2 is aligned to 8 and has place for yuvsrc2[-1] and yuvsrc2[width]
    yuvsrc3 = (uint32_t *) ( ((uintptr_t) yuvsrc2) + 4 * width + 8); // yuvsrc2 is aligned to 8 and has place for yuvsrc2[-1] and yuvsrc2[width]
    yuvsrc4 = (uint32_t *) ( ((uintptr_t) yuvsrc3) + 4 * width + 8); // yuvsrc2 is aligned to 8 and has place for yuvsrc2[-1] and yuvsrc2[width]

    patternsrc1 = (uint8_t *) ( ((uintptr_t) yuvsrc4) + 4 * width + 8); // patternsrc1 is aligned to 8 and has place for patternsrc1[-1] and patternsrc1[width]

    tmpdst1 = (uint16_t *) ( ((uintptr_t) patternsrc1) + width + 8); // tmpdst1 is aligned to 8
    tmpdst2 = (uint16_t *) ( ((uintptr_t) tmpdst1    ) + 8 * width); // tmpdst2 is aligned to 8

    yuvpalette = (uint32_t *) ( ((uintptr_t) tmpdst2) + 8 * width); // yuvpalette is aligned to 8

    blendTable = (_blendTable *) ( ((uintptr_t) yuvpalette) + 4 * 256); // blendTable is aligned to 8

    prepare_blend_table(blendTable, (uint32_t *) &(blendTable[12 + 1 + 1 + 1]), width);

    prepare_yuv_palette(rgbpalette, yuvpalette);

    // wait for copying first line to tmpsrc1
    syscalls.qdma_wait(QDMA_0);

    // start copying second line to tmpsrc2
    syscalls.qdma_copy1d(QDMA_0, tmpsrc2, src + srcstride, width);
    src += 2*srcstride;

    // convert first line to yuv and rgb
    convert_yuv_rgb_8(tmpsrc1, yuvsrc1, rgbsrc1, yuvpalette, rgbpalette, width);

    // wait for copying second line to tmpsrc2
    syscalls.qdma_wait(QDMA_0);

    // start copying third line to tmpsrc3
    syscalls.qdma_copy1d(QDMA_0, tmpsrc3, src, width);
    src += srcstride;

    // convert second line to yuv and rgb
    convert_yuv_rgb_8(tmpsrc2, yuvsrc2, rgbsrc2, yuvpalette, rgbpalette, width);

    // calculate pattern for first line
    calculate_pattern(yuvsrc1, yuvsrc1, yuvsrc2, patternsrc1, width);

    // calculate and process blend table 1 for first line - upper left corner
    fill_blendtable1(patternsrc1, &(hqTable[0]), blendTable, width);

    fill_dst1(rgbsrc1, rgbsrc1, rgbsrc2, tmpdst1, blendTable);

    // calculate and process blend table 2 for first line - upper right corner
    fill_blendtable2(patternsrc1, &(hqTable[256]), blendTable, width);

    fill_dst2(rgbsrc1, rgbsrc1, rgbsrc2, tmpdst1+1, blendTable);

    // calculate and process blend table 3 for first line - lower right corner
    fill_blendtable3(patternsrc1, &(hqTable[512]), blendTable, width);

    fill_dst3(rgbsrc1, rgbsrc1, rgbsrc2, tmpdst1 + 2*width + 1, blendTable);

    // calculate and process blend table 4 for first line - lower left corner
    fill_blendtable4(patternsrc1, &(hqTable[768]), blendTable, width);

    fill_dst4(rgbsrc1, rgbsrc1, rgbsrc2, tmpdst1 + 2*width, blendTable);

    // start copying first two destination lines from tmpdst1
    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst1, 4*width, 4*width, 2);
    dst += dststride;

    for (height -= 3; height != 0; height--)
    {
        // wait for copying "third" line to tmpsrc3
        syscalls.qdma_wait(QDMA_0);

        // convert "third" line to yuv and rgb
        convert_yuv_rgb_8(tmpsrc3, yuvsrc3, rgbsrc3, yuvpalette, rgbpalette, width);

        // calculate pattern for "second" line
        calculate_pattern(yuvsrc1, yuvsrc2, yuvsrc3, patternsrc1, width);

        // calculate and process blend table 1 for "second" line - upper left corner
        fill_blendtable1(patternsrc1, &(hqTable[0]), blendTable, width);

        fill_dst1(rgbsrc1, rgbsrc2, rgbsrc3, tmpdst2, blendTable);

        // calculate and process blend table 2 for "second" line - upper right corner
        fill_blendtable2(patternsrc1, &(hqTable[256]), blendTable, width);

        fill_dst2(rgbsrc1, rgbsrc2, rgbsrc3, tmpdst2+1, blendTable);

        // start copying "fourth" line to tmpsrc4
        syscalls.qdma_copy1d(QDMA_0, tmpsrc4, src, width);
        src += srcstride;

        // calculate and process blend table 3 for "second" line - lower right corner
        fill_blendtable3(patternsrc1, &(hqTable[512]), blendTable, width);

        fill_dst3(rgbsrc1, rgbsrc2, rgbsrc3, tmpdst2 + 2*width + 1, blendTable);

        // calculate and process blend table 4 for "second" line - lower left corner
        fill_blendtable4(patternsrc1, &(hqTable[768]), blendTable, width);

        fill_dst4(rgbsrc1, rgbsrc2, rgbsrc3, tmpdst2 + 2*width, blendTable);

        // wait for copying "first" two destination lines from tmpdst1
        syscalls.qdma_wait(QDMA_1);

        // start copying "second" two destination lines from tmpdst2
        syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);
        dst += dststride;

        // "rename" buffers
        {
            uint16_t * restrict tmpdst;
            tmpdst = tmpdst1;
            tmpdst1 = tmpdst2;
            tmpdst2 = tmpdst;
        }
        {
            uint8_t * restrict tmpsrc;
            tmpsrc = tmpsrc1;
            tmpsrc1 = tmpsrc2;
            tmpsrc2 = tmpsrc3;
            tmpsrc3 = tmpsrc4;
            tmpsrc4 = tmpsrc;
        }
        {
            uint16_t * restrict rgbsrc;
            rgbsrc = rgbsrc1;
            rgbsrc1 = rgbsrc2;
            rgbsrc2 = rgbsrc3;
            rgbsrc3 = rgbsrc4;
            rgbsrc4 = rgbsrc;
        }
        {
            uint32_t * restrict yuvsrc;
            yuvsrc = yuvsrc1;
            yuvsrc1 = yuvsrc2;
            yuvsrc2 = yuvsrc3;
            yuvsrc3 = yuvsrc4;
            yuvsrc4 = yuvsrc;
        }
    }

    // last-1 line:

    // wait for copying last line to tmpsrc3
    syscalls.qdma_wait(QDMA_0);

    // convert last line to yuv and rgb
    convert_yuv_rgb_8(tmpsrc3, yuvsrc3, rgbsrc3, yuvpalette, rgbpalette, width);

    // calculate pattern for last line
    calculate_pattern(yuvsrc1, yuvsrc2, yuvsrc3, patternsrc1, width);

    // calculate and process blend table 1 for last-1 line - upper left corner
    fill_blendtable1(patternsrc1, &(hqTable[0]), blendTable, width);

    fill_dst1(rgbsrc1, rgbsrc2, rgbsrc3, tmpdst2, blendTable);

    // calculate and process blend table 2 for last-1 line - upper right corner
    fill_blendtable2(patternsrc1, &(hqTable[256]), blendTable, width);

    fill_dst2(rgbsrc1, rgbsrc2, rgbsrc3, tmpdst2+1, blendTable);

    // calculate and process blend table 3 for last-1 line - lower right corner
    fill_blendtable3(patternsrc1, &(hqTable[512]), blendTable, width);

    fill_dst3(rgbsrc1, rgbsrc2, rgbsrc3, tmpdst2 + 2*width + 1, blendTable);

    // calculate and process blend table 4 for last-1 line - lower left corner
    fill_blendtable4(patternsrc1, &(hqTable[768]), blendTable, width);

    fill_dst4(rgbsrc1, rgbsrc2, rgbsrc3, tmpdst2 + 2*width, blendTable);

    // wait for copying last-2 two destination lines from tmpdst1
    syscalls.qdma_wait(QDMA_1);

    // start copying last-1 two destination lines from tmpdst2
    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);
    dst += dststride;

    // last line:

    // calculate pattern for last line
    calculate_pattern(yuvsrc2, yuvsrc3, yuvsrc3, patternsrc1, width);

    // calculate and process blend table 1 for last line - upper left corner
    fill_blendtable1(patternsrc1, &(hqTable[0]), blendTable, width);

    fill_dst1(rgbsrc2, rgbsrc3, rgbsrc3, tmpdst1, blendTable);

    // calculate and process blend table 2 for last line - upper right corner
    fill_blendtable2(patternsrc1, &(hqTable[256]), blendTable, width);

    fill_dst2(rgbsrc2, rgbsrc3, rgbsrc3, tmpdst1+1, blendTable);

    // calculate and process blend table 3 for last line - lower right corner
    fill_blendtable3(patternsrc1, &(hqTable[512]), blendTable, width);

    fill_dst3(rgbsrc2, rgbsrc3, rgbsrc3, tmpdst1 + 2*width + 1, blendTable);

    // calculate and process blend table 4 for last line - lower left corner
    fill_blendtable4(patternsrc1, &(hqTable[768]), blendTable, width);

    fill_dst4(rgbsrc2, rgbsrc3, rgbsrc3, tmpdst1 + 2*width, blendTable);

    // start copying last two destination lines from tmpdst1
    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst1, 4*width, 4*width, 2);
    //dst += dststride;

    // wait for copying last-1 two destination lines from tmpdst2
    syscalls.qdma_wait(QDMA_1);

    // wait for copying last two destination lines from tmpdst1
    syscalls.qdma_wait(QDMA_0);
}

void _hq2x_16_16(const uint16_t * restrict src, uint16_t * restrict dst, int width, int srcstride, int dststride, int height, uint16_t *hqTable)
{
    _nassert((int)src % 8 == 0);
    _nassert((int)dst % 8 == 0);
    _nassert(width % 8 == 0);
    _nassert(width >= 32);
    _nassert(width <= MAX_WIDTH);
    _nassert(srcstride >= 2 * width);
    _nassert(dststride >= 4 * width);
    _nassert(srcstride % 8 == 0);
    _nassert(dststride % 8 == 0);
    _nassert(height >= 4);
    _nassert((int)hqTable % 8 == 0);

    // hqTable should be in L1D SRAM

    uint16_t * restrict tmpsrc1;
    uint16_t * restrict tmpsrc2;
    uint16_t * restrict tmpsrc3;
    uint16_t * restrict tmpsrc4;

    uint32_t * restrict yuvsrc1;
    uint32_t * restrict yuvsrc2;
    uint32_t * restrict yuvsrc3;
    uint32_t * restrict yuvsrc4;

    uint8_t * restrict patternsrc1;

    uint16_t * restrict tmpdst1;
    uint16_t * restrict tmpdst2;

    _blendTable * restrict blendTable;


    tmpsrc1 = (uint16_t *) (DSP_L1DSRAM_SCRATCHBUFFER_BASE_DSP + 8); // tmpsrc1 is aligned to 8 and has place for tmpsrc1[-1] and tmpsrc1[width]

    // start copying first line to tmpsrc1
    syscalls.qdma_copy1d(QDMA_0, tmpsrc1, src, 2*width);

    tmpsrc2 = (uint16_t *) ( ((uintptr_t) tmpsrc1) + 2 * width + 8); // tmpsrc2 is aligned to 8 and has place for tmpsrc2[-1] and tmpsrc2[width]
    tmpsrc3 = (uint16_t *) ( ((uintptr_t) tmpsrc2) + 2 * width + 8); // tmpsrc3 is aligned to 8 and has place for tmpsrc3[-1] and tmpsrc3[width]
    tmpsrc4 = (uint16_t *) ( ((uintptr_t) tmpsrc3) + 2 * width + 8); // tmpsrc4 is aligned to 8 and has place for tmpsrc4[-1] and tmpsrc4[width]

    yuvsrc1 = (uint32_t *) ( ((uintptr_t) tmpsrc4) + 2 * width + 8); // yuvsrc1 is aligned to 8 and has place for yuvsrc1[-1] and yuvsrc1[width]
    yuvsrc2 = (uint32_t *) ( ((uintptr_t) yuvsrc1) + 4 * width + 8); // yuvsrc2 is aligned to 8 and has place for yuvsrc2[-1] and yuvsrc2[width]
    yuvsrc3 = (uint32_t *) ( ((uintptr_t) yuvsrc2) + 4 * width + 8); // yuvsrc2 is aligned to 8 and has place for yuvsrc2[-1] and yuvsrc2[width]
    yuvsrc4 = (uint32_t *) ( ((uintptr_t) yuvsrc3) + 4 * width + 8); // yuvsrc2 is aligned to 8 and has place for yuvsrc2[-1] and yuvsrc2[width]

    patternsrc1 = (uint8_t *) ( ((uintptr_t) yuvsrc4) + 4 * width + 8); // patternsrc1 is aligned to 8 and has place for patternsrc1[-1] and patternsrc1[width]

    tmpdst1 = (uint16_t *) ( ((uintptr_t) patternsrc1) + width + 8); // tmpdst1 is aligned to 8
    tmpdst2 = (uint16_t *) ( ((uintptr_t) tmpdst1    ) + 8 * width); // tmpdst2 is aligned to 8

    blendTable = (_blendTable *) ( ((uintptr_t) tmpdst2) + 8 * width); // blendTable is aligned to 8

    prepare_blend_table(blendTable, (uint32_t *) &(blendTable[12 + 1 + 1 + 1]), width);

    // wait for copying first line to tmpsrc1
    syscalls.qdma_wait(QDMA_0);

    // start copying second line to tmpsrc2
    syscalls.qdma_copy1d(QDMA_0, tmpsrc2, src + srcstride/2, 2*width);
    src += srcstride;

    // convert first line to yuv
    convert_yuv_16(tmpsrc1, yuvsrc1, width);

    // wait for copying second line to tmpsrc2
    syscalls.qdma_wait(QDMA_0);

    // start copying third line to tmpsrc3
    syscalls.qdma_copy1d(QDMA_0, tmpsrc3, src, 2*width);
    src += srcstride/2;

    // convert second line to yuv
    convert_yuv_16(tmpsrc2, yuvsrc2, width);

    // calculate pattern for first line
    calculate_pattern(yuvsrc1, yuvsrc1, yuvsrc2, patternsrc1, width);

    // calculate and process blend table 1 for first line - upper left corner
    fill_blendtable1(patternsrc1, &(hqTable[0]), blendTable, width);

    fill_dst1(tmpsrc1, tmpsrc1, tmpsrc2, tmpdst1, blendTable);

    // calculate and process blend table 2 for first line - upper right corner
    fill_blendtable2(patternsrc1, &(hqTable[256]), blendTable, width);

    fill_dst2(tmpsrc1, tmpsrc1, tmpsrc2, tmpdst1+1, blendTable);

    // calculate and process blend table 3 for first line - lower right corner
    fill_blendtable3(patternsrc1, &(hqTable[512]), blendTable, width);

    fill_dst3(tmpsrc1, tmpsrc1, tmpsrc2, tmpdst1 + 2*width + 1, blendTable);

    // calculate and process blend table 4 for first line - lower left corner
    fill_blendtable4(patternsrc1, &(hqTable[768]), blendTable, width);

    fill_dst4(tmpsrc1, tmpsrc1, tmpsrc2, tmpdst1 + 2*width, blendTable);

    // start copying first two destination lines from tmpdst1
    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst1, 4*width, 4*width, 2);
    dst += dststride;

    for (height -= 3; height != 0; height--)
    {
        // wait for copying "third" line to tmpsrc3
        syscalls.qdma_wait(QDMA_0);

        // convert "third" line to yuv
        convert_yuv_16(tmpsrc3, yuvsrc3, width);

        // calculate pattern for "second" line
        calculate_pattern(yuvsrc1, yuvsrc2, yuvsrc3, patternsrc1, width);

        // calculate and process blend table 1 for "second" line - upper left corner
        fill_blendtable1(patternsrc1, &(hqTable[0]), blendTable, width);

        fill_dst1(tmpsrc1, tmpsrc2, tmpsrc3, tmpdst2, blendTable);

        // calculate and process blend table 2 for "second" line - upper right corner
        fill_blendtable2(patternsrc1, &(hqTable[256]), blendTable, width);

        fill_dst2(tmpsrc1, tmpsrc2, tmpsrc3, tmpdst2+1, blendTable);

        // start copying "fourth" line to tmpsrc4
        syscalls.qdma_copy1d(QDMA_0, tmpsrc4, src, 2*width);
        src += srcstride/2;

        // calculate and process blend table 3 for "second" line - lower right corner
        fill_blendtable3(patternsrc1, &(hqTable[512]), blendTable, width);

        fill_dst3(tmpsrc1, tmpsrc2, tmpsrc3, tmpdst2 + 2*width + 1, blendTable);

        // calculate and process blend table 4 for "second" line - lower left corner
        fill_blendtable4(patternsrc1, &(hqTable[768]), blendTable, width);

        fill_dst4(tmpsrc1, tmpsrc2, tmpsrc3, tmpdst2 + 2*width, blendTable);

        // wait for copying "first" two destination lines from tmpdst1
        syscalls.qdma_wait(QDMA_1);

        // start copying "second" two destination lines from tmpdst2
        syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);
        dst += dststride;

        // "rename" buffers
        {
            uint16_t * restrict tmpdst;
            tmpdst = tmpdst1;
            tmpdst1 = tmpdst2;
            tmpdst2 = tmpdst;
        }
        {
            uint16_t * restrict tmpsrc;
            tmpsrc = tmpsrc1;
            tmpsrc1 = tmpsrc2;
            tmpsrc2 = tmpsrc3;
            tmpsrc3 = tmpsrc4;
            tmpsrc4 = tmpsrc;
        }
        {
            uint32_t * restrict yuvsrc;
            yuvsrc = yuvsrc1;
            yuvsrc1 = yuvsrc2;
            yuvsrc2 = yuvsrc3;
            yuvsrc3 = yuvsrc4;
            yuvsrc4 = yuvsrc;
        }
    }

    // last-1 line:

    // wait for copying last line to tmpsrc3
    syscalls.qdma_wait(QDMA_0);

    // convert last line to yuv
    convert_yuv_16(tmpsrc3, yuvsrc3, width);

    // calculate pattern for last-1 line
    calculate_pattern(yuvsrc1, yuvsrc2, yuvsrc3, patternsrc1, width);

    // calculate and process blend table 1 for last-1 line - upper left corner
    fill_blendtable1(patternsrc1, &(hqTable[0]), blendTable, width);

    fill_dst1(tmpsrc1, tmpsrc2, tmpsrc3, tmpdst2, blendTable);

    // calculate and process blend table 2 for last-1 line - upper right corner
    fill_blendtable2(patternsrc1, &(hqTable[256]), blendTable, width);

    fill_dst2(tmpsrc1, tmpsrc2, tmpsrc3, tmpdst2+1, blendTable);

    // calculate and process blend table 3 for last-1 line - lower right corner
    fill_blendtable3(patternsrc1, &(hqTable[512]), blendTable, width);

    fill_dst3(tmpsrc1, tmpsrc2, tmpsrc3, tmpdst2 + 2*width + 1, blendTable);

    // calculate and process blend table 4 for last-1 line - lower left corner
    fill_blendtable4(patternsrc1, &(hqTable[768]), blendTable, width);

    fill_dst4(tmpsrc1, tmpsrc2, tmpsrc3, tmpdst2 + 2*width, blendTable);

    // wait for copying last-2 two destination lines from tmpdst1
    syscalls.qdma_wait(QDMA_1);

    // start copying last-1 two destination lines from tmpdst2
    syscalls.qdma_copy2d(QDMA_1, dst, dststride, tmpdst2, 4*width, 4*width, 2);
    dst += dststride;

    // last line:

    // calculate pattern for last line
    calculate_pattern(yuvsrc2, yuvsrc3, yuvsrc3, patternsrc1, width);

    // calculate and process blend table 1 for last line - upper left corner
    fill_blendtable1(patternsrc1, &(hqTable[0]), blendTable, width);

    fill_dst1(tmpsrc2, tmpsrc3, tmpsrc3, tmpdst1, blendTable);

    // calculate and process blend table 2 for last line - upper right corner
    fill_blendtable2(patternsrc1, &(hqTable[256]), blendTable, width);

    fill_dst2(tmpsrc2, tmpsrc3, tmpsrc3, tmpdst1+1, blendTable);

    // calculate and process blend table 3 for last line - lower right corner
    fill_blendtable3(patternsrc1, &(hqTable[512]), blendTable, width);

    fill_dst3(tmpsrc2, tmpsrc3, tmpsrc3, tmpdst1 + 2*width + 1, blendTable);

    // calculate and process blend table 4 for last line - lower left corner
    fill_blendtable4(patternsrc1, &(hqTable[768]), blendTable, width);

    fill_dst4(tmpsrc2, tmpsrc3, tmpsrc3, tmpdst1 + 2*width, blendTable);

    // start copying last two destination lines from tmpdst1
    syscalls.qdma_copy2d(QDMA_0, dst, dststride, tmpdst1, 4*width, 4*width, 2);
    //dst += dststride;

    // wait for copying last-1 two destination lines from tmpdst2
    syscalls.qdma_wait(QDMA_1);

    // wait for copying last two destination lines from tmpdst1
    syscalls.qdma_wait(QDMA_0);
}

